from os import PathLike
from pathlib import Path

import numpy as np
from scipy.interpolate import SmoothBivariateSpline

import openmdao.api as om


class GeomorphologyGridData:
    """
    A class to represent gridded geomorphology data for a given wind farm site
    domain.

    Represents either bathymetry data for offshore sites or topography data for
    onshore sites.
    """

    # alias for meshed data, and promote dimension to 2
    x_data = np.atleast_2d([0.0])  # x location in km
    y_data = np.atleast_2d([0.0])  # y location in km
    z_data = np.atleast_2d([0.0])  # depth in m

    # alias for meshed material data, promote dimension to 2
    x_material_data = np.atleast_2d([0.0])  # x location in km of material datapoint
    y_material_data = np.atleast_2d([0.0])  # y location in km of material datapoint
    material_data = np.atleast_2d(["soil"])  # bed material at each point

    sea_level = 0.0  # sea level in m

    _interpolator_device = None  # placeholder for interpolator (for depth evaluation)

    def check_valid_geomorphology(self):
        assert self.x_data.ndim == 2, "data must be 2D"  # make sure it's 2D first

        assert np.all(
            self.x_data.shape == self.y_data.shape
        ), "x and y data must be the same shape"
        assert np.all(
            self.x_data.shape == self.z_data.shape
        ), "x and depth data must be the same shape"

        return True

    def check_valid_material(self):
        assert (
            self.x_material_data.ndim == 2
        ), "data must be 2D"  # make sure it's 2D first

        assert np.all(
            self.x_material_data.shape == self.y_material_data.shape
        ), "x and y material data must be the same shape"
        assert np.all(self.x_material_data.shape == self.material_data.shape) or (
            self.material_data.size == 1
        ), "x and material data must be the same shape or material data must be a singleton"

        return True

    def get_shape(self):
        """
        Get the shape of the geomorphology data.

        Returns
        -------
        tuple
            The shape of the geomorphology data.
        """

        self.check_valid_geomorphology()  # ensure that the current data is valid
        return self.z_data.shape  # data shape

    def get_material_shape(self):
        """
        Get the shape of the material data.

        Returns
        -------
        tuple
            The shape of the material data.
        """

        self.check_valid_material()  # ensure that the current data is valid
        return self.material_data.shape  # data shape

    def set_data_values(
        self,
        x_data_in,
        y_data_in,
        z_data_in,
    ):
        """
        Set the values of the geomorphology data.

        Parameters
        ----------
        x_data_in : np.ndarray
            A 2D numpy array indicating the x-dimension locations of the points.
        y_data_in : np.ndarray
            A 2D numpy array indicating the y-dimension locations of the points.
        z_data_in : np.ndarray
            A 2D numpy array indicating the depth at each point.
        material_data_in : np.ndarray, optional
            A 2D numpy array indicating the bed material at each point.
        """

        # set the values that are handed in
        self.x_data = x_data_in.copy()
        self.y_data = y_data_in.copy()
        self.z_data = z_data_in.copy()

        self.check_valid_geomorphology()  # ensure that the input data is valid

    def set_material_values(
        self,
        x_material_data_in,
        y_material_data_in,
        material_data_in,
    ):
        """
        Set the values of the material data.

        Parameters
        ----------
        x_material_data_in : np.ndarray
            A 2D numpy array indicating the x-dimension locations of the points.
        y_material_data_in : np.ndarray
            A 2D numpy array indicating the y-dimension locations of the points.
        material_data_in : np.ndarray
            A 2D numpy array indicating the bed material at each point.
        """

        # set the values that are handed in
        self.x_material_data = x_material_data_in.copy()
        self.y_material_data = y_material_data_in.copy()
        self.material_data = material_data_in.copy()

        self.check_valid_material()  # ensure that the input data is valid

    def get_z_data(self):
        """Get the depth at a given location."""
        return self.z_data

    def get_material_data(self):
        """Get the material data at a given location."""
        return self.material_data

    def evaluate(
        self,
        x_query,
        y_query,
        return_derivs=False,
        interp_method="spline",
    ):
        """
        Evaluate the depth at a given location.

        Parameters
        ----------
        x_query : np.array
            The x locations to sample in km
        y_query : np.array
            The y locations to sample in km

        Returns
        -------
        np.array
            the depth at the given locations if return_derivs is False
        tuple
            the derivatives if return_derivs is True
        """

        x_query = np.atleast_1d(x_query)  # ensure x_query is a 1D array
        y_query = np.atleast_1d(y_query)  # ensure y_query is a 1D array

        if interp_method == "spline":
            # or, smooth bivariate spline from scipy implementation

            if self._interpolator_device is None:
                # create and alias the interpolator object
                interpolator_sbs = self._interpolator_device = SmoothBivariateSpline(
                    self.x_data.flatten(),
                    self.y_data.flatten(),
                    self.z_data.flatten(),
                    bbox=[
                        np.min(self.x_data),
                        np.max(self.x_data),
                        np.min(self.y_data),
                        np.max(self.y_data),
                    ],
                )
            else:
                # assert the interpolator is of the smoothbivariate spline type or its parent class
                assert isinstance(
                    self._interpolator_device, SmoothBivariateSpline
                ), "interpolator must be a SmoothBivariateSpline"
                # alias
                interpolator_sbs = self._interpolator_device

            # make interpolation
            if return_derivs:
                # and if desired, take its derivatives
                dz_dx = interpolator_sbs(x_query, y_query, grid=False, dx=1, dy=0)
                dz_dy = interpolator_sbs(x_query, y_query, grid=False, dx=0, dy=1)
                return (dz_dx, dz_dy)  # and return
            else:
                z_query = interpolator_sbs(x_query, y_query, grid=False)
                return z_query  # just return

        else:
            raise NotImplementedError(
                f"{interp_method} interpolation scheme for evaluate not implemented yet. -cfrontin"
            )


class BathymetryGridData(GeomorphologyGridData):
    """
    A class to represent gridded bathymetry data for a given wind farm site
    domain.

    Represents the bathymetry data for offshore sites. Can be used for floating
    mooring system anchors or for fixed-bottom foundations. Should specialize
    geomorphology data for bathymetry-specific considerations.
    """

    def load_moorpy_soil(self, file_soil: PathLike):
        """
        Load soil data from a MoorPy soil file.

        Experimental: reader may not be able to read validly formatted file in
        in the presence of unanticipated comments, whitespace, etc.

        Parameters
        ----------
        file_soil : PathLike
            The path to the soil data file
        """

        # create placholder objects in function local scope
        grid_soil = None
        x_coord = None
        y_coord = None

        with open(file_soil, "r") as f_soil:
            idx_y = 0  # indexer for y coordinate as file is read

            # iterate over lines in the soil file
            for idx_line, line in enumerate(f_soil.readlines()):

                if idx_line == 0:  # moorpy header line must be first
                    assert line.startswith("--- MoorPy Soil Input File ---")
                    continue
                if idx_line == 1:  # next line defines the grid size in x
                    assert line.startswith("nGridX")  # guarantee this is the case
                    nGridX = int(line.split()[1])  # extract the number
                    x_coord = np.zeros((nGridX,))  # prepare a coord array
                    continue

    def load_moorpy_bathymetry(self, file_bathymetry: PathLike):
        """
        Load bathymetry data from a MoorPy bathymetry grid file.

        Experimental: reader may not be able to read validly formatted file in
        in the presence of unanticipated comments, whitespace, etc.

        Parameters
        ----------
        file_bathymetry : str
            The path to the bathymetry data file
        """

        # create placeholder objects in function local scope
        grid_bathy = None
        x_coord = None
        y_coord = None

        with open(file_bathymetry, "r") as f_bathy:
            idx_y = 0  # indexer for y coordinate as file is read

            # iterate over lines in the bathymetry file
            for idx_line, line in enumerate(f_bathy.readlines()):

                if idx_line == 0:  # moorpy header line must be first
                    assert line.startswith("--- MoorPy Bathymetry Input File ---")
                    continue
                if idx_line == 1:  # next line defines the grid size in x
                    assert line.startswith("nGridX")  # guarantee this is the case
                    nGridX = int(line.split()[1])  # extract the number
                    x_coord = np.zeros((nGridX,))  # prepare a coord array
                    continue
                if idx_line == 2:  # next line defines the grid size in y
                    assert line.startswith("nGridY")  # guarantee this is the case
                    nGridY = int(line.split()[1])  # extract the number
                    y_coord = np.zeros((nGridY,))  # prepare a coord array
                    grid_bathy = np.zeros((nGridX, nGridY))  # prepare a grid
                    continue

                if idx_line == 3:  # next line should define the x coordinates
                    x_coord_tgt = [float(x) for x in line.split()]  # extract
                    assert len(x_coord_tgt) == nGridX  # verify length
                    x_coord = np.array(x_coord_tgt)  # convert to array
                    continue

                if (
                    idx_line > 3
                ):  # all other lines should be y coordinate then gridpoint data
                    if not line.strip():
                        continue  # if the line is empty or whitespace, skip it

                    y_coord_tgt = float(line.split()[0])  # extract the y coordinate
                    bathy_row_tgt = [
                        float(b) for b in line.split()[1:]
                    ]  # extract the bathymetry data
                    assert len(bathy_row_tgt) == nGridX  # verify length
                    y_coord[idx_y] = y_coord_tgt  # set the y coordinate
                    grid_bathy[:, idx_y] = bathy_row_tgt  # set the bathymetry data
                    idx_y += 1  # increment the y indexer
            assert idx_y == nGridY  # verify that all y coordinates were read

        # save into the geomorphology data object
        self.y_data, self.x_data = np.meshgrid(y_coord, x_coord)
        self.z_data = grid_bathy

        self.check_valid_geomorphology()  # make sure the loaded file is legit before exiting


class TopographyGridData(GeomorphologyGridData):
    """
    A class to represent gridded terrain data for a given wind farm site domain.

    Represents the terrain data for onshore sites. Should specialize
    geomorphology data for topography-specific considerations.
    """

    pass



import warnings
import numpy as np

import openmdao.api as om
from wisdem.plant_financese.plant_finance import PlantFinance as PlantFinance_orig
from wisdem.landbosse.landbosse_omdao.landbosse import LandBOSSE as LandBOSSE_orig
from wisdem.orbit.orbit_api import Orbit as Orbit_orig

from ard.cost.approximate_turbine_spacing import SpacingApproximations


class LandBOSSEWithSpacingApproximations(om.Group):
    """
    OpenMDAO group that connects the SpacingApproximations component to the LandBOSSE component.

    This group calculates the turbine spacing using the SpacingApproximations and passes it
    to the LandBOSSE component for further cost estimation.
    """

    def initialize(self):
        """Initialize the group and declare options."""
        self.options.declare(
            "modeling_options", types=dict, desc="Ard modeling options"
        )

    def setup(self):
        """Set up the group by adding and connecting components."""
        # Add the PrimarySpacingApproximations component
        self.add_subsystem(
            "spacing_approximations",
            SpacingApproximations(modeling_options=self.options["modeling_options"]),
            promotes_inputs=["total_length_cables"],
        )

        # Add the LandBOSSE component
        self.add_subsystem(
            "landbosse",
            LandBOSSEGroup(modeling_options=self.options["modeling_options"]),
            promotes_inputs=[
                "*",
                (
                    "turbine_spacing_rotor_diameters",
                    "internal_turbine_spacing_rotor_diameters",
                ),
                (
                    "row_spacing_rotor_diameters",
                    "internal_row_spacing_rotor_diameters",
                ),
            ],
            promotes_outputs=["*"],  # Expose all outputs from LandBOSSE
        )

        # Connect the turbine and row spacing outputs from the approximations to LandBOSSE
        self.connect(
            "spacing_approximations.primary_turbine_spacing_diameters",
            "internal_turbine_spacing_rotor_diameters",
        )

        self.connect(
            "spacing_approximations.secondary_turbine_spacing_diameters",
            "internal_row_spacing_rotor_diameters",
        )


class LandBOSSEGroup(om.Group):

    def initialize(self):
        """Initialize the group and declare options."""
        self.options.declare(
            "modeling_options", types=dict, desc="Ard modeling options"
        )

    def setup(self):

        # add IVCs for landbosse
        variable_mapping = LandBOSSE_setup_latents(
            modeling_options=self.options["modeling_options"]
        )

        # create source independent variable components for LandBOSSE inputs
        for key, meta in variable_mapping.items():
            if key in ["num_turbines", "number_of_blades"]:
                comp = om.IndepVarComp()
                comp.add_discrete_output(name=key, val=meta["val"])
                self.add_subsystem(f"IVC_landbosse_{key}", comp, promotes=["*"])
            else:
                self.add_subsystem(
                    f"IVC_landbosse_{key}",
                    om.IndepVarComp(key, val=meta["val"], units=meta["units"]),
                    promotes=["*"],
                )

        # add landbosse
        self.add_subsystem(
            "landbosse",
            LandBOSSE_orig(),
            promotes=[
                "total_capex",
                "total_capex_kW",
                "bos_capex_kW",
                "turbine_spacing_rotor_diameters",
                "row_spacing_rotor_diameters",
            ],
        )

        # connect
        for key, val in variable_mapping.items():
            self.connect(key, f"landbosse.{key}")


class ORBITGroup(om.Group):

    def initialize(self):
        """Initialize the group and declare options."""
        self.options.declare(
            "modeling_options", types=dict, desc="Ard modeling options"
        )

    def setup(self):

        # add IVCs for landbosse
        variable_mapping = ORBIT_setup_latents(
            modeling_options=self.options["modeling_options"]
        )

        # create source independent variable components for LandBOSSE inputs
        for key, meta in variable_mapping.items():
            if key in ["number_of_turbines", "number_of_blades", "num_mooring_lines"]:
                comp = om.IndepVarComp()
                comp.add_discrete_output(name=key, val=meta["val"])
                self.add_subsystem(f"IVC_orbit_{key}", comp, promotes=["*"])
            else:
                self.add_subsystem(
                    f"IVC_orbit_{key}",
                    om.IndepVarComp(key, val=meta["val"], units=meta["units"]),
                    promotes=["*"],
                )

        # add orbit
        self.add_subsystem(
            "orbit",
            Orbit_orig(
                floating=self.options["modeling_options"]["floating"],
                jacket=self.options["modeling_options"].get("jacket"),
                jacket_legs=self.options["modeling_options"].get("jacket_legs"),
            ),
            promotes=[
                "total_capex",
                "total_capex_kW",
                "bos_capex",
                "installation_capex",
                "plant_turbine_spacing",
                "plant_row_spacing",
            ],
        )

        # connect
        for key, val in variable_mapping.items():
            self.connect(key, f"orbit.{key}")


class FinanceSEGroup(om.Group):

    def initialize(self):
        """Initialize the group and declare options."""
        self.options.declare(
            "modeling_options", types=dict, desc="Ard modeling options"
        )

    def setup(self):

        # add IVCs for landbosse
        variable_mapping = FinanceSE_setup_latents(
            modeling_options=self.options["modeling_options"]
        )

        # create source independent variable components for LandBOSSE inputs
        for key, meta in variable_mapping.items():
            if key in [
                "turbine_number",
            ]:
                comp = om.IndepVarComp()
                comp.add_discrete_output(name=key, val=meta["val"])
                self.add_subsystem(f"IVC_financese_{key}", comp, promotes=["*"])
            else:
                self.add_subsystem(
                    f"IVC_financese_{key}",
                    om.IndepVarComp(key, val=meta["val"], units=meta["units"]),
                    promotes=["*"],
                )

        # add financese #TODO check promotes
        self.add_subsystem(
            "financese",
            PlantFinance_orig(),
            promotes=[
                "offset_tcc_per_kW",
                "plant_aep_in",
                "bos_per_kW",
                # "tcc_per_kW",
                "lcoe",
            ],
        )

        # connect
        for key, val in variable_mapping.items():
            self.connect(key, f"financese.{key}")


class TurbineCapitalCosts(om.ExplicitComponent):
    """
    A simple component to compute the turbine capital costs.

    Inputs
    ------
    machine_rating : float
        rating of the wind turbine in kW
    tcc_per_kW : float
        turbine capital costs per kW (as output from WISDEM tools)
    offset_tcc_per_kW : float
        additional tcc per kW (offset)

    Discrete Inputs
    ---------------
    turbine_number : int
        number of turbines in the farm

    Outputs
    -------
    tcc : float
        turbine capital costs in USD
    """

    def setup(self):
        """Setup of OM component."""
        self.add_input("machine_rating", 0.0, units="kW")
        self.add_input("tcc_per_kW", 0.0, units="USD/kW")
        self.add_input("offset_tcc_per_kW", 0.0, units="USD/kW")
        self.add_discrete_input("turbine_number", 0)
        self.add_output("tcc", 0.0, units="USD")

    def setup_partials(self):
        """Derivative setup for OM component."""
        # complex step for simple gradients
        self.declare_partials("*", "*", method="cs")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """Computation for the OM compoent."""
        # Unpack parameters
        t_rating = inputs["machine_rating"]
        n_turbine = discrete_inputs["turbine_number"]
        tcc_per_kW = inputs["tcc_per_kW"] + inputs["offset_tcc_per_kW"]
        outputs["tcc"] = n_turbine * tcc_per_kW * t_rating


class OperatingExpenses(om.ExplicitComponent):
    """
    A simple component to compute the operating costs.

    Inputs
    ------
    machine_rating : float
        rating of the wind turbine in kW
    opex_per_kW : float
        annual operating and maintenance costs per kW (as output from WISDEM
        tools)

    Discrete Inputs
    ---------------
    turbine_number : int
        number of turbines in the farm

    Outputs
    -------
    opex : float
        annual operating and maintenance costs in USD
    """

    def setup(self):
        """Setup of OM component."""
        self.add_input("machine_rating", 0.0, units="kW")
        self.add_input("opex_per_kW", 0.0, units="USD/kW/yr")
        self.add_discrete_input("turbine_number", 0)
        self.add_output("opex", 0.0, units="USD/yr")

    def setup_partials(self):
        """Derivative setup for OM component."""
        # complex step for simple gradients
        self.declare_partials("*", "*", method="cs")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """Computation for the OM compoent."""
        # Unpack parameters
        t_rating = inputs["machine_rating"]
        n_turbine = discrete_inputs["turbine_number"]
        opex_per_kW = inputs["opex_per_kW"]
        outputs["opex"] = n_turbine * opex_per_kW * t_rating


def LandBOSSE_setup_latents(modeling_options: dict) -> None:
    """
    A function to set up the LandBOSSE latent variables using modeling options.

    Parameters
    ----------
    prob : openmdao.api.Problem
        an OpenMDAO problem for which we want to setup the LandBOSSE latent
        variables
    modeling_options : dict
        a modeling options dictionary
    """

    # Define the mapping between OpenMDAO variable names and modeling_options keys
    offshore_fixed_keys = [
        "monopile_mass",
        "monopile_cost",
    ]

    offshore_floating_keys = [
        "num_mooring_lines",
        "mooring_line_mass",
        "mooring_line_diameter",
        "mooring_line_length",
        "anchor_mass",
        "floating_substructure_cost",
    ]

    def _base_common():
        return {
            "num_turbines": {
                "val": modeling_options["layout"]["N_turbines"],
                "units": None,
            },
            "turbine_rating_MW": {
                "val": modeling_options["windIO_plant"]["wind_farm"]["turbine"][
                    "performance"
                ]["rated_power"]
                / 1.0e6,
                "units": "MW",
            },
            "hub_height_meters": {
                "val": modeling_options["windIO_plant"]["wind_farm"]["turbine"][
                    "hub_height"
                ],
                "units": "m",
            },
            "rotor_diameter_m": {
                "val": modeling_options["windIO_plant"]["wind_farm"]["turbine"][
                    "rotor_diameter"
                ],
                "units": "m",
            },
            "number_of_blades": {
                "val": modeling_options["costs"]["num_blades"],
                "units": None,
            },
            "tower_mass": {
                "val": modeling_options["costs"]["tower_mass"],
                "units": "t",
            },
            "nacelle_mass": {
                "val": modeling_options["costs"]["nacelle_mass"],
                "units": "t",
            },
            "blade_mass": {
                "val": modeling_options["costs"]["blade_mass"],
                "units": "t",
            },
            "commissioning_cost_kW": {
                "val": modeling_options["costs"]["commissioning_cost_kW"],
                "units": "USD/kW",
            },
            "decommissioning_cost_kW": {
                "val": modeling_options["costs"]["decommissioning_cost_kW"],
                "units": "USD/kW",
            },
        }

    if any(key in modeling_options["costs"] for key in offshore_fixed_keys):
        variable_mapping = _base_common()
        variable_mapping.update(
            {
                "monopile_mass": {
                    "val": modeling_options["costs"]["monopile_mass"],
                    "units": "kg",
                },
                "monopile_cost": {
                    "val": modeling_options["costs"]["monopile_cost"],
                    "units": "USD",
                },
            }
        )
    elif any(key in modeling_options["costs"] for key in offshore_floating_keys):
        variable_mapping = _base_common()
        variable_mapping.update(
            {
                "num_mooring_lines": {
                    "val": modeling_options["costs"]["num_mooring_lines"],
                    "units": None,
                },
                "mooring_line_mass": {
                    "val": modeling_options["costs"]["mooring_line_mass"],
                    "units": "kg",
                },
                "mooring_line_diameter": {
                    "val": modeling_options["costs"]["mooring_line_diameter"],
                    "units": "m",
                },
                "mooring_line_length": {
                    "val": modeling_options["costs"]["mooring_line_length"],
                    "units": "m",
                },
                "anchor_mass": {
                    "val": modeling_options["costs"]["anchor_mass"],
                    "units": "kg",
                },
                "floating_substructure_cost": {
                    "val": modeling_options["costs"]["floating_substructure_cost"],
                    "units": "USD",
                },
            }
        )
    else:
        variable_mapping = _base_common()
        variable_mapping.update(
            {
                "rated_thrust_N": {
                    "val": modeling_options["costs"]["rated_thrust_N"],
                    "units": "N",
                },
                "gust_velocity_m_per_s": {
                    "val": modeling_options["costs"]["gust_velocity_m_per_s"],
                    "units": "m/s",
                },
                "blade_surface_area": {
                    "val": modeling_options["costs"]["blade_surface_area"],
                    "units": "m**2",
                },
                "hub_mass": {
                    "val": modeling_options["costs"]["hub_mass"],
                    "units": "kg",
                },
                "foundation_height": {
                    "val": modeling_options["costs"]["foundation_height"],
                    "units": "m",
                },
                "trench_len_to_substation_km": {
                    "val": modeling_options["costs"]["trench_len_to_substation_km"],
                    "units": "km",
                },
                "distance_to_interconnect_mi": {
                    "val": modeling_options["costs"]["distance_to_interconnect_mi"],
                    "units": "mi",
                },
                "interconnect_voltage_kV": {
                    "val": modeling_options["costs"]["interconnect_voltage_kV"],
                    "units": "kV",
                },
            }
        )

    return variable_mapping


def ORBIT_setup_latents(modeling_options: dict) -> None:
    """
    A function to set up the ORBIT latent variables using modeling options.

    Parameters
    ----------
    prob : openmdao.api.Problem
        an OpenMDAO problem for which we want to setup the ORBIT latent
        variables
    modeling_options : dict
        a modeling options dictionary
    """

    variable_mapping = {
        "turbine_rating": {
            "val": modeling_options["windIO_plant"]["wind_farm"]["turbine"][
                "performance"
            ]["rated_power"],
            "units": "W",
        },
        "site_depth": {"val": modeling_options["site_depth"], "units": "m"},
        "number_of_turbines": {
            "val": modeling_options["layout"]["N_turbines"],
            "units": None,
        },
        "number_of_blades": {
            "val": modeling_options["costs"]["num_blades"],
            "units": None,
        },
        "hub_height": {
            "val": modeling_options["windIO_plant"]["wind_farm"]["turbine"][
                "hub_height"
            ],
            "units": "m",
        },
        "turbine_rotor_diameter": {
            "val": modeling_options["windIO_plant"]["wind_farm"]["turbine"][
                "rotor_diameter"
            ],
            "units": "m",
        },
        "tower_length": {
            "val": modeling_options["costs"]["tower_length"],
            "units": "m",
        },
        "tower_mass": {"val": modeling_options["costs"]["tower_mass"], "units": "t"},
        "nacelle_mass": {
            "val": modeling_options["costs"]["nacelle_mass"],
            "units": "t",
        },
        "blade_mass": {"val": modeling_options["costs"]["blade_mass"], "units": "t"},
        "turbine_capex": {
            "val": modeling_options["costs"]["turbine_capex"],
            "units": "USD/kW",
        },
        "site_mean_windspeed": {
            "val": modeling_options["costs"]["site_mean_windspeed"],
            "units": "m/s",
        },
        "turbine_rated_windspeed": {
            "val": modeling_options["costs"]["turbine_rated_windspeed"],
            "units": "m/s",
        },
        "commissioning_cost_kW": {
            "val": modeling_options["costs"]["commissioning_cost_kW"],
            "units": "USD/kW",
        },
        "decommissioning_cost_kW": {
            "val": modeling_options["costs"]["decommissioning_cost_kW"],
            "units": "USD/kW",
        },
        "plant_substation_distance": {
            "val": modeling_options["costs"]["plant_substation_distance"],
            "units": "km",
        },
        "interconnection_distance": {
            "val": modeling_options["costs"]["interconnection_distance"],
            "units": "km",
        },
        "site_distance": {
            "val": modeling_options["costs"]["site_distance"],
            "units": "km",
        },
        "site_distance_to_landfall": {
            "val": modeling_options["costs"]["site_distance_to_landfall"],
            "units": "km",
        },
        "port_cost_per_month": {
            "val": modeling_options["costs"]["port_cost_per_month"],
            "units": "USD/month",
        },
        "construction_insurance": {
            "val": modeling_options["costs"]["construction_insurance"],
            "units": "USD/kW",
        },
        "construction_financing": {
            "val": modeling_options["costs"]["construction_financing"],
            "units": "USD/kW",
        },
        "contingency": {
            "val": modeling_options["costs"]["contingency"],
            "units": "USD/kW",
        },
        "site_auction_price": {
            "val": modeling_options["costs"]["site_auction_price"],
            "units": "USD",
        },
        "site_assessment_cost": {
            "val": modeling_options["costs"]["site_assessment_cost"],
            "units": "USD",
        },
        "construction_plan_cost": {
            "val": modeling_options["costs"]["construction_plan_cost"],
            "units": "USD",
        },
        "installation_plan_cost": {
            "val": modeling_options["costs"]["installation_plan_cost"],
            "units": "USD",
        },
        "boem_review_cost": {
            "val": modeling_options["costs"]["boem_review_cost"],
            "units": "USD",
        },
    }

    # Add floating-foundation specific keys if applicable
    if modeling_options["floating"]:
        variable_mapping.update(
            {
                "num_mooring_lines": {
                    "val": modeling_options["costs"]["num_mooring_lines"],
                    "units": None,
                },
                "mooring_line_mass": {
                    "val": modeling_options["costs"]["mooring_line_mass"],
                    "units": "kg",
                },
                "mooring_line_diameter": {
                    "val": modeling_options["costs"]["mooring_line_diameter"],
                    "units": "m",
                },
                "mooring_line_length": {
                    "val": modeling_options["costs"]["mooring_line_length"],
                    "units": "m",
                },
                "anchor_mass": {
                    "val": modeling_options["costs"]["anchor_mass"],
                    "units": "kg",
                },
                "transition_piece_mass": {
                    "val": modeling_options["costs"]["transition_piece_mass"],
                    "units": "t",
                },
                "transition_piece_cost": {
                    "val": modeling_options["costs"]["transition_piece_cost"],
                    "units": "USD",
                },
                "floating_substructure_cost": {
                    "val": modeling_options["costs"]["floating_substructure_cost"],
                    "units": "USD",
                },
            }
        )
    # Add fixed-foundation (mooring) specific keys if applicable
    else:
        variable_mapping.update(
            {
                "monopile_mass": {
                    "val": modeling_options["costs"]["monopile_mass"],
                    "units": "t",
                },
                "monopile_cost": {
                    "val": modeling_options["costs"]["monopile_cost"],
                    "units": "USD",
                },
                "monopile_length": {
                    "val": modeling_options["costs"]["monopile_length"],
                    "units": "m",
                },
                "monopile_diameter": {
                    "val": modeling_options["costs"]["monopile_diameter"],
                    "units": "m",
                },
                "transition_piece_mass": {
                    "val": modeling_options["costs"]["transition_piece_mass"],
                    "units": "t",
                },
                "transition_piece_cost": {
                    "val": modeling_options["costs"]["transition_piece_cost"],
                    "units": "USD",
                },
            }
        )
    # TODO include jacket-type foundation
    # # if jacket
    # prob.set_val(
    #     comp2promotion_map["orbit.orbit.jacket_r_foot"],
    #     modeling_options["turbine"]["costs"]["jacket_r_foot"])
    # prob.set_val(
    #     comp2promotion_map["orbit.orbit.jacket_length"],
    #     modeling_options["turbine"]["costs"]["jacket_length"])
    # prob.set_val(
    #     comp2promotion_map["orbit.orbit.jacket_mass"],
    #     modeling_options["turbine"]["costs"]["jacket_mass"])
    # prob.set_val(
    #     comp2promotion_map["orbit.orbit.jacket_cost"],
    #     modeling_options["turbine"]["costs"]["jacket_cost"])
    # prob.set_val(
    #     comp2promotion_map["orbit.orbit.transition_piece_mass"],
    #     modeling_options["turbine"]["costs"]["transition_piece_mass"])
    # prob.set_val(
    #     comp2promotion_map["orbit.orbit.transition_piece_cost"],
    #     modeling_options["turbine"]["costs"]["transition_piece_cost"])

    return variable_mapping


def FinanceSE_setup_latents(modeling_options):
    """
    A function to set up the FinanceSE latent variables using modeling options.

    Parameters
    ----------
    prob : openmdao.api.Problem
        an OpenMDAO problem for which we want to setup the FinanceSE latent
        variables
    modeling_options : dict
        a modeling options dictionary
    """

    # Define the mapping between OpenMDAO variable names and modeling_options keys
    variable_mapping = {
        "turbine_number": {
            "val": int(modeling_options["layout"]["N_turbines"]),
            "units": None,
        },
        "machine_rating": {
            "val": modeling_options["windIO_plant"]["wind_farm"]["turbine"][
                "performance"
            ]["rated_power"],
            "units": "W",
        },
        "tcc_per_kW": {
            "val": modeling_options["costs"]["tcc_per_kW"],
            "units": "USD/kW",
        },
        "opex_per_kW": {
            "val": modeling_options["costs"]["opex_per_kW"],
            "units": "USD/kW/year",
        },
    }

    return variable_mapping



from pathlib import Path
import copy

import numpy as np

import floris
import floris.turbine_library.turbine_utilities

import ard.utils.logging as ard_logging
import ard.farm_aero.templates as templates


def create_FLORIS_turbine_from_windIO(
    windIOplant: dict,
    modeling_options: dict = {},
) -> dict:
    """
    takes a windIO plant specification and creates a FLORIS turbine dictionary

    Parameters
    ----------
    windIOplant : dict
        a full, presumed validated, windIO plant specification file
    modeling_options : dict, optional
        an Ard modeling options dictionary, which can contain FLORIS options for
        configuring the FLORIS turbine

    Returns
    -------
    dict
        the FLORIS dictionary that was created, with references deep copied

    Raises
    ------
    NotImplementedError
        if a valid windIO configuration for which reconciliation with the
        FLORIS specification has not been implemented yet
    IndexError
        if the windIO file is in an apparently invalid state
    """

    # extract the turbine... assuming a single one for now
    windIOturbine = windIOplant["wind_farm"]["turbine"]

    # generate dictionary for FLORIS utility
    tdd = {}
    wind_speeds_Ct_curve = windIOturbine["performance"]["Ct_curve"]["Ct_wind_speeds"]
    values_Ct_curve = windIOturbine["performance"]["Ct_curve"]["Ct_values"]
    if "Cp_curve" in windIOturbine["performance"]:
        wind_speeds_Cp_curve = windIOturbine["performance"]["Cp_curve"][
            "Cp_wind_speeds"
        ]
        values_Cp_curve = windIOturbine["performance"]["Cp_curve"]["Cp_values"]
        if not np.allclose(wind_speeds_Cp_curve, wind_speeds_Ct_curve):
            raise NotImplementedError(
                "Ct and Cp curves are specified with different indep. variables. "
                "windIO allows seperate indep. variables for power/thrust curve "
                "specification, while FLORIS requires common indep. variables. "
                "Reconciliation has not been implemented yet."
            )
        tdd["wind_speed"] = wind_speeds_Cp_curve
        tdd["power_coefficient"] = values_Cp_curve
        tdd["thrust_coefficient"] = values_Ct_curve
    elif "power_curve" in windIOturbine["performance"]:
        wind_speeds_power_curve = windIOturbine["performance"]["power_curve"][
            "power_wind_speeds"
        ]
        values_power_curve = windIOturbine["performance"]["power_curve"]["power_values"]
        if not np.allclose(wind_speeds_power_curve, wind_speeds_Ct_curve):
            raise NotImplementedError(
                "Ct and power curves are specified with different indep. variables. "
                "windIO allows seperate indep. variables for power/thrust curve "
                "specification, while FLORIS requires common indep. variables. "
                "Reconciliation has not been implemented yet."
            )
        tdd["wind_speed"] = wind_speeds_power_curve
        tdd["power"] = [
            v / 1.0e3 for v in values_power_curve
        ]  # windIO data comes in in W and FLORIS takes kW
        tdd["thrust_coefficient"] = values_Ct_curve
    elif all(
        val in windIOturbine["performance"]
        for val in [
            "rated_power",
            "rated_wind_speed",
            "cutin_wind_speed",
            "cutout_wind_speed",
        ]
    ):
        # extract key values
        rated_power = windIOturbine["rated_power"]
        rated_wind_speed = windIOturbine["rated_wind_speed"]
        cutin_wind_speed = windIOturbine["cutin_wind_speed"]
        cutout_wind_speed = windIOturbine["cutout_wind_speed"]

        # ### SYNTHETIC POWER CURVE ###
        # for this section, we assume synthetic behavior based on the values
        # specified in a windIO plant turbine specification, based on region:
        #
        # - region I: before cut-in, $V < `cutin_wind_speed`$
        #       $P(V) = 0$
        # - region II: harvest all available power, $`cutin_wind_speed` < V < `rated_wind_speed`$
        #       $P(V) = 0.5 * rho * A * V^3 * C_{P,\max}$
        #   for $P_{\mathrm{rated}} = 0.5 * rho * A * V_{\mathrm{rated}}^3 * C_{P,\max}$,
        #   it follows that $P(V) = P_{\mathrm{rated}} \\left( \frac{V}{V_{\mathrm{rated}}} \\right)^3$
        # - region III: throttle power to rated, $`rated_wind_speed` < V < `cutout_wind_speed`$
        #       $P(V) = P_{\mathrm{rated}}$
        # - region IV: after cut-out, $V > `cutout_wind_speed`$
        #       $P(V) = 0$

        # compute based on reg. I, II, III, IV of synthetic power curve
        values_power_curve = (
            wind_speeds_Ct_curve**3 / rated_wind_speed**3 * rated_power
        )  # scales proportionally in reg. II
        values_power_curve[wind_speeds_Ct_curve >= rated_wind_speed] = (
            rated_power  # flat in reg. III
        )
        values_power_curve[wind_speeds_Ct_curve <= cutin_wind_speed] = (
            0.0  # zero in reg. I
        )
        values_power_curve[wind_speeds_Ct_curve >= cutout_wind_speed] = (
            0.0  # zero in reg. IV
        )

        # pack and ship
        tdd["wind_speed"] = wind_speeds_Ct_curve
        tdd["power"] = (
            values_power_curve / 1e3
        )  # windIO data comes in in W and FLORIS takes kW
        tdd["thrust_coefficient"] = values_Ct_curve
    else:
        raise IndexError(
            "The windIO file appears to be invalid. Try validating and re-running."
        )

    # create a FLORIS turbine using the utility
    turbine_FLORIS = (
        floris.turbine_library.turbine_utilities.build_cosine_loss_turbine_dict(
            turbine_data_dict=tdd,
            turbine_name=windIOturbine["name"],
            hub_height=windIOturbine["hub_height"],
            rotor_diameter=windIOturbine["rotor_diameter"],
            TSR=windIOturbine.get("TSR"),
            generator_efficiency=windIOturbine.get("generator_efficiency", 1.0),
            ref_air_density=modeling_options.get("floris", {}).get(
                "ref_air_density", 1.225
            ),
        )
    )

    # append peak shaving reduction fraction and TI threshhold they exist
    psf_val = modeling_options.get("floris", {}).get("peak_shaving_fraction")
    psf_thresh = modeling_options.get("floris", {}).get(
        "peak_shaving_TI_threshold", 0.0
    )
    if psf_val is not None:
        # if they exist, set them
        turbine_FLORIS["power_thrust_table"]["peak_shaving_fraction"] = psf_val
        turbine_FLORIS["power_thrust_table"]["peak_shaving_TI_threshold"] = psf_thresh

    # # If an export filename is given, write it out
    # if filename_turbine_FLORIS is not None:
    #     with open(filename_turbine_FLORIS, "w") as file_turbine_FLORIS:
    #         yaml.safe_dump(turbine_FLORIS, file_turbine_FLORIS)

    return copy.deepcopy(turbine_FLORIS)


class FLORISFarmComponent:
    """
    Secondary-inherit component for managing FLORIS for farm simulations.

    This is a base class for farm aerodynamics simulations using FLORIS, which
    should cover all the necessary configuration, reproducibility config file
    saving, and output directory management.

    It is not a child class of an OpenMDAO components, but it is designed to
    mirror the form of the OM component, so that FLORIS activities are separated
    to have run times that correspond to the similarly-named OM component
    methods. It is intended to be a second-inherit base class for FLORIS-based
    OpenMDAO components, and will not work unless the calling object is a
    specialized class that _also_ specializes `openmdao.api.Component`.

    Options
    -------
    case_title : str
        a "title" for the case, used to disambiguate runs in practice
    """

    def initialize(self):
        """Initialization-time FLORIS management."""
        self.options.declare("case_title")

    @ard_logging.component_log_capture
    def setup(self):
        """Setup-time FLORIS management."""

        # set up FLORIS
        self.fmodel = floris.FlorisModel("defaults")
        data_path = self.options["data_path"]
        self.fmodel.set(
            turbine_type=[
                create_FLORIS_turbine_from_windIO(self.windIO, self.modeling_options),
            ],
            wind_shear=self.windIO["site"]["energy_resource"]["wind_resource"].get(
                "shear"
            ),
            reference_wind_height=getattr(
                self.wind_query,
                "reference_height",
                None,
            ),
        )

        self.case_title = self.options["case_title"]
        self.dir_floris = ard_logging.get_storage_directory(
            self, "inputs", get_iter=True, clean=True
        )

    def compute(self, inputs):
        """
        Compute-time FLORIS management.

        Compute-time FLORIS management should be specialized based on use case.
        If the base class is not specialized, an error will be raised.
        """

        raise NotImplementedError("compute must be specialized,")

    def setup_partials(self):
        """Derivative setup for OM component."""
        # for FLORIS, no derivatives. use FD because FLORIS is cheap
        self.declare_partials("*", "*", method="fd")

    def get_AEP_farm(self):
        """Get the AEP of a FLORIS farm."""
        return self.fmodel.get_farm_AEP()

    def get_power_farm(self):
        """Get the farm power of a FLORIS farm at each wind condition."""
        return self.fmodel.get_farm_power()

    def get_power_turbines(self):
        """Get the turbine powers of a FLORIS farm at each wind condition."""
        return self.fmodel.get_turbine_powers().T

    def get_thrust_turbines(self):
        """Get the turbine thrusts of a FLORIS farm at each wind condition."""
        # FLORIS computes the thrust precursors, compute and return thrust
        # use pure FLORIS to get these values for consistency

        # prepare to unpack thrust data that is not pre-computed in FLORIS
        CT_turbines = self.fmodel.get_turbine_thrust_coefficients()
        V_turbines = self.fmodel.turbine_average_velocities
        rho_floris = self.fmodel.core.flow_field.air_density
        A_floris = np.pi * self.fmodel.core.farm.rotor_diameters**2 / 4

        # unpacking procedure
        # from FLORIS's floris_model.py:564 at a6fc5d35aa32614edc450dc399c42af60a816887
        thrust_turbines = CT_turbines * (0.5 * rho_floris * A_floris * V_turbines**2)
        if isinstance(self.fmodel.wind_data, floris.WindRose) or isinstance(
            self.fmodel.wind_data, floris.WindRoseWRG
        ):
            thrust_turbines_densified = np.full(
                (
                    np.prod(self.fmodel.wind_data.freq_table.shape),
                    self.fmodel.core.farm.n_turbines,
                ),
                0.0,
            )
            thrust_turbines_densified[self.fmodel.wind_data.non_zero_freq_mask, :] = (
                thrust_turbines
            )
            return thrust_turbines_densified.T
        else:
            return thrust_turbines.T

    def dump_floris_yamlfile(self, dir_output=None):
        """
        Export the current FLORIS inputs to a YAML file file for reproducibility of the analysis.
        The file will be saved in the `dir_output` directory, or in the current working directory
        if `dir_output` is None.
        """
        if dir_output is None:
            dir_output = self.dir_floris
        self.fmodel.core.to_file(Path(dir_output, "batch.yaml"))


class FLORISBatchPower(templates.BatchFarmPowerTemplate, FLORISFarmComponent):
    """
    Component class for computing a batch power analysis using FLORIS.

    A component class that evaluates a series of farm power and associated
    quantities using FLORIS. Inherits the interface from
    `templates.BatchFarmPowerTemplate` and the computational guts from
    `FLORISFarmComponent`.

    Options
    -------
    case_title : str
        a "title" for the case, used to disambiguate runs in practice (inherited
        from `FLORISFarmComponent`)
    modeling_options : dict
        a modeling options dictionary (inherited via
        `templates.BatchFarmPowerTemplate`)
    wind_query : floris.wind_data.WindRose
        a WindQuery objects that specifies the wind conditions that are to be
        computed (inherited from `templates.BatchFarmPowerTemplate`)

    Inputs
    ------
    x_turbines : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the turbines,
        with length `N_turbines` (inherited via
        `templates.BatchFarmPowerTemplate`)
    y_turbines : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the turbines,
        with length `N_turbines` (inherited via
        `templates.BatchFarmPowerTemplate`)
    yaw_turbines : np.ndarray
        a numpy array indicating the yaw angle to drive each turbine to with
        respect to the ambient wind direction, with length `N_turbines`
        (inherited via `templates.BatchFarmPowerTemplate`)

    Outputs
    -------
    power_farm : np.ndarray
        an array of the farm power for each of the wind conditions that have
        been queried (inherited from `templates.BatchFarmPowerTemplate`)
    power_turbines : np.ndarray
        an array of the farm power for each of the turbines in the farm across
        all of the conditions that have been queried on the wind rose
        (`N_turbines`, `N_wind_conditions`) (inherited from
        `templates.BatchFarmPowerTemplate`)
    thrust_turbines : np.ndarray
        an array of the wind turbine thrust for each of the turbines in the farm
        across all of the conditions that have been queried on the wind rose
        (`N_turbines`, `N_wind_conditions`) (inherited from
        `templates.BatchFarmPowerTemplate`)
    """

    def initialize(self):
        super().initialize()  # run super class script first!
        FLORISFarmComponent.initialize(self)  # FLORIS superclass

    @ard_logging.component_log_capture
    def setup(self):
        super().setup()  # run super class script first!
        FLORISFarmComponent.setup(self)  # setup a FLORIS run

    @ard_logging.component_log_capture
    def setup_partials(self):
        FLORISFarmComponent.setup_partials(self)

    @ard_logging.component_log_capture
    def compute(self, inputs, outputs):

        # generate the list of conditions for evaluation
        self.time_series = floris.TimeSeries(
            wind_directions=np.degrees(np.array(self.wind_query.wind_directions)),
            wind_speeds=np.array(self.wind_query.wind_speeds),
            turbulence_intensities=np.array(self.wind_query.turbulence_intensities),
        )

        # set up and run the floris model
        self.fmodel.set(
            layout_x=inputs["x_turbines"],
            layout_y=inputs["y_turbines"],
            wind_data=self.time_series,
            yaw_angles=np.array([inputs["yaw_turbines"]]),
            reference_wind_height=getattr(
                self.wind_query,
                "reference_height",
                None,
            ),
        )
        if "peak_shaving_fraction" in self.modeling_options.get("floris", {}):
            self.fmodel.set_operation_model("peak-shaving")

        self.fmodel.run()

        # dump the yaml to re-run this case on demand
        # FLORISFarmComponent.dump_floris_yamlfile(self, self.dir_floris)

        # FLORIS computes the powers
        outputs["AEP_farm"] = FLORISFarmComponent.get_AEP_farm(self)
        outputs["power_farm"] = FLORISFarmComponent.get_power_farm(self)
        if self.options["modeling_options"]["aero"]["return_turbine_output"]:
            outputs["power_turbines"] = FLORISFarmComponent.get_power_turbines(self)
            outputs["thrust_turbines"] = FLORISFarmComponent.get_thrust_turbines(self)


class FLORISAEP(templates.FarmAEPTemplate):
    """
    Component class for computing an AEP analysis using FLORIS.

    A component class that evaluates a series of farm power and associated
    quantities using FLORIS with a wind rose to make an AEP estimate. Inherits
    the interface from `templates.FarmAEPTemplate` and the computational guts
    from `FLORISFarmComponent`.

    Options
    -------
    case_title : str
        a "title" for the case, used to disambiguate runs in practice (inherited
        from `FLORISFarmComponent`)
    modeling_options : dict
        a modeling options dictionary (inherited via
        `templates.FarmAEPTemplate`)
    wind_query : floris.wind_data.WindRose
        a WindQuery objects that specifies the wind conditions that are to be
        computed (inherited from `templates.FarmAEPTemplate`)

    Inputs
    ------
    x_turbines : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the turbines,
        with length `N_turbines` (inherited via `templates.FarmAEPTemplate`)
    y_turbines : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the turbines,
        with length `N_turbines` (inherited via `templates.FarmAEPTemplate`)
    yaw_turbines : np.ndarray
        a numpy array indicating the yaw angle to drive each turbine to with
        respect to the ambient wind direction, with length `N_turbines`
        (inherited via `templates.FarmAEPTemplate`)

    Outputs
    -------
    AEP_farm : float
        the AEP of the farm given by the analysis (inherited from
        `templates.FarmAEPTemplate`)
    power_farm : np.ndarray
        an array of the farm power for each of the wind conditions that have
        been queried (inherited from `templates.FarmAEPTemplate`)
    power_turbines : np.ndarray
        an array of the farm power for each of the turbines in the farm across
        all of the conditions that have been queried on the wind rose
        (`N_turbines`, `N_wind_conditions`) (inherited from
        `templates.FarmAEPTemplate`)
    thrust_turbines : np.ndarray
        an array of the wind turbine thrust for each of the turbines in the farm
        across all of the conditions that have been queried on the wind rose
        (`N_turbines`, `N_wind_conditions`) (inherited from
        `templates.FarmAEPTemplate`)
    """

    def initialize(self):
        super().initialize()  # run super class script first!
        FLORISFarmComponent.initialize(self)  # add on FLORIS superclass

    @ard_logging.component_log_capture
    def setup(self):
        super().setup()  # run super class script first!
        FLORISFarmComponent.setup(self)  # setup a FLORIS run

    @ard_logging.component_log_capture
    def setup_partials(self):
        super().setup_partials()

    @ard_logging.component_log_capture
    def compute(self, inputs, outputs):

        # set up and run the floris model
        self.fmodel.set(
            layout_x=inputs["x_turbines"],
            layout_y=inputs["y_turbines"],
            wind_data=self.wind_query,
            yaw_angles=np.array([inputs["yaw_turbines"]]),
            reference_wind_height=getattr(
                self.wind_query,
                "reference_height",
                None,
            ),
        )
        if "peak_shaving_fraction" in self.modeling_options.get("floris", {}):
            self.fmodel.set_operation_model("peak-shaving")

        self.fmodel.run()

        # dump the yaml to re-run this case on demand
        FLORISFarmComponent.dump_floris_yamlfile(self, self.dir_floris)

        # FLORIS computes the powers
        outputs["AEP_farm"] = FLORISFarmComponent.get_AEP_farm(self)
        outputs["power_farm"] = FLORISFarmComponent.get_power_farm(self)
        outputs["power_turbines"] = FLORISFarmComponent.get_power_turbines(self)
        outputs["thrust_turbines"] = FLORISFarmComponent.get_thrust_turbines(self)

    @ard_logging.component_log_capture
    def setup_partials(self):
        FLORISFarmComponent.setup_partials(self)



import numpy as np

from shapely.geometry import Point
from shapely.ops import unary_union

import openmdao.api as om

_R_earth = 6371008.8  # Earth radius, m


def calculate_viewshed_section_angle(
    D_rotor: float,  # rotor diameter, m
    h_hub: float,  # hub height, m
    R_earth: float = _R_earth,  # Earth radius, m
    h_terrain: float = 0.0,  # mean height of prevailing terrain wrt turbine base, m
) -> float:  # returns arc angle (rad) of viewshed
    H = D_rotor / 2 + h_hub - h_terrain
    return np.arccos(R_earth / (R_earth + H))


def calculate_viewshed_arc_length(
    D_rotor: float,  # rotor diameter, m
    h_hub: float,  # hub height, m
    R_earth: float = _R_earth,  # Earth radius, m
    h_terrain: float = 0.0,  # mean height of prevailing terrain wrt turbine base, m
) -> float:  # returns arc length of viewshed
    angle_arc = calculate_viewshed_section_angle(
        D_rotor,
        h_hub,
        R_earth=R_earth,
        h_terrain=h_terrain,
    )
    return R_earth * angle_arc


def calculate_viewshed_arc_length_smallangle(
    D_rotor: float,  # rotor diameter, m
    h_hub: float,  # hub height, m
    R_earth: float = _R_earth,  # Earth radius, m
    h_terrain: float = 0.0,  # mean height of prevailing terrain wrt turbine base, m
) -> float:  # returns arc length of viewshed
    angle_arc = calculate_viewshed_section_angle(
        D_rotor,
        h_hub,
        R_earth=R_earth,
        h_terrain=h_terrain,
    )
    return R_earth * np.sin(angle_arc)


class ViewshedAreaComp(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        self.modeling_options = self.options["modeling_options"]
        self.windIO = self.modeling_options["windIO_plant"]

        self.D_rotor = self.windIO["wind_farm"]["turbine"]["rotor_diameter"]
        self.h_hub = self.windIO["wind_farm"]["turbine"]["hub_height"]
        self.N_turbines = self.modeling_options["layout"]["N_turbines"]

        self.add_input("x_turbines", np.zeros((self.N_turbines,)), units="m")
        self.add_input("y_turbines", np.zeros((self.N_turbines,)), units="m")
        # self.add_output('viewshed_arc_length', val=0.0, units='m', desc='Viewshed arc length')
        self.add_output("area_viewshed", val=0.0, units="km**2", desc="Viewshed area")

    def setup_partials(self):
        # declare FD because no derivative is available
        self.declare_partials("area_viewshed", "*", method="fd")

    def compute(self, inputs, outputs):

        # get the single-turbine viewshed arc length
        D_rotor = self.D_rotor
        h_hub = self.h_hub
        h_terrain = 0.0

        # project onto 2D surface plane
        x_turbines = inputs["x_turbines"]
        y_turbines = inputs["y_turbines"]
        D_rotor_turbines = D_rotor * np.ones_like(x_turbines)
        h_hub_turbines = h_hub * np.ones_like(x_turbines)
        R_viewshed_turbines = calculate_viewshed_arc_length(
            D_rotor_turbines,
            h_hub_turbines,
            h_terrain=h_terrain,
        )

        # create a list of shapely circles
        viewshed_circles = [
            Point(x, y).buffer(r)
            for x, y, r in zip(
                x_turbines.flatten(),
                y_turbines.flatten(),
                R_viewshed_turbines.flatten(),
            )
        ]

        # compute the union of all circles
        viewshed_union = unary_union(viewshed_circles)

        # calculate the area of the union in square kilometers
        viewshed_union_area_km2 = viewshed_union.area / 1e6

        # pack and send output
        outputs["area_viewshed"] = viewshed_union_area_km2



import numpy as np
import jax.numpy as jnp
import jax
import ard.utils.mathematics
import openmdao.api as om


class TurbineSpacing(om.ExplicitComponent):
    """
    A class to return distances between each pair of turbines without duplicates.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary (inherited from `FarmAeroTemplate`)

    Inputs
    ------
    x_turbines : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the turbines,
        with length `N_turbines` (mirrored w.r.t. `FarmAeroTemplate`)
    y_turbines : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the turbines,
        with length `N_turbines` (mirrored w.r.t. `FarmAeroTemplate`)
    turbine_spacing : np.ndarray
        a 1D numpy array indicating the distances between turbines,
        with length (N_turbines - 1)*N_turbines/2. where, for 3 turbines, turbine_spacing[0]
        is the distance between turbines 0 and 1, turbine_spacing[1] is the distance between
        turbines 0 and 2, and turbine_spacing[2] is the distance between turbines 1 and 2.
        The array is the flattened upper-triangular portion of the distance matrix.
    """

    def initialize(self):
        """Initialization of the OpenMDAO component."""
        self.options.declare("modeling_options")

    def setup(self):
        """Setup of the OpenMDAO component."""

        # load modeling options
        self.modeling_options = self.options["modeling_options"]
        self.N_turbines = int(self.modeling_options["layout"]["N_turbines"])
        self.N_distances = int((self.N_turbines - 1) * self.N_turbines / 2)

        # set up inputs and outputs for mooring system
        self.add_input(
            "x_turbines", jnp.zeros((self.N_turbines,)), units="km"
        )  # x location of the mooring platform in km w.r.t. reference coordinates
        self.add_input(
            "y_turbines", jnp.zeros((self.N_turbines,)), units="km"
        )  # y location of the mooring platform in km w.r.t. reference coordinates

        self.add_output(
            "turbine_spacing",
            jnp.zeros(self.N_distances),
            units="km",
        )

    def setup_partials(self):
        """Derivative setup for the OpenMDAO component."""
        # the default (but not preferred!) derivatives are FDM
        self.declare_partials("*", "*", method="exact")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """Computation for the OpenMDAO component."""

        # unpack the working variables
        x_turbines = inputs["x_turbines"]
        y_turbines = inputs["y_turbines"]

        spacing_distances = calculate_turbine_spacing(x_turbines, y_turbines)

        outputs["turbine_spacing"] = spacing_distances

    def compute_partials(self, inputs, partials, discrete_inputs=None):

        # unpack the working variables
        x_turbines = inputs["x_turbines"]
        y_turbines = inputs["y_turbines"]

        jacobian = calculate_turbine_spacing_jac(x_turbines, y_turbines)

        partials["turbine_spacing", "x_turbines"] = jacobian[0]
        partials["turbine_spacing", "y_turbines"] = jacobian[1]


def calculate_turbine_spacing(
    x_turbines: np.ndarray,
    y_turbines: np.ndarray,
) -> np.ndarray:
    """Calculate the spacing between every pair of turbines with no duplicates.

    Args:
        x_turbines (np.ndarray): a 1D numpy array indicating the x-dimension locations of the turbines,
            with length `N_turbines`.
        y_turbines (np.ndarray): a 1D numpy array indicating the y-dimension locations of the turbines,
            with length `N_turbines`.

    Returns:
        turbine distances (np.ndarray): a 1D numpy array indicating the distances between turbines
            with length (N_turbines - 1)*N_turbines/2. The array is the flattened
            upper-triangular portion of the distance matrix. For 3 turbines, turbine_spacing[0] is the
            distance between turbines 0 and 1, turbine_spacing[1] is the distance between
            turbines 0 and 2, and turbine_spacing[2] is the distance between turbines 1 and 2.
    """

    N_turbines = len(x_turbines)

    # Create index pairs for i < j (upper triangle without diagonal)
    i_indices, j_indices = jnp.triu_indices(N_turbines, k=1)

    # Compute deltas
    dx = x_turbines[j_indices] - x_turbines[i_indices]
    dy = y_turbines[j_indices] - y_turbines[i_indices]
    deltas = jnp.stack([dx, dy], axis=1)

    # Vectorized norm calculation
    spacing_distance = ard.utils.mathematics.smooth_norm_vec(deltas)

    return spacing_distance


calculate_turbine_spacing = jax.jit(calculate_turbine_spacing)
calculate_turbine_spacing_jac = jax.jacrev(calculate_turbine_spacing, argnums=[0, 1])



from warnings import warn

import networkx as nx
import numpy as np

from optiwindnet.mesh import make_planar_embedding
from optiwindnet.interarraylib import L_from_site
from optiwindnet.heuristics import EW_presolver
from optiwindnet.MILP import OWNWarmupFailed, solver_factory, ModelOptions

from . import templates


def _own_L_from_inputs(inputs: dict, discrete_inputs: dict) -> nx.Graph:
    # get the metadata and data for the OWN warm-starter from the inputs
    T = len(inputs["x_turbines"])
    R = len(inputs["x_substations"])
    name_case = "farm"
    if discrete_inputs["x_border"] is not None:
        B = len(discrete_inputs["x_border"])
    else:
        B = 0
    VertexC = np.empty((R + T + B, 2), dtype=float)
    VertexC[:T, 0] = inputs["x_turbines"]
    VertexC[:T, 1] = inputs["y_turbines"]
    VertexC[-R:, 0] = inputs["x_substations"]
    VertexC[-R:, 1] = inputs["y_substations"]

    # add perturbation to duplicate turbine/substation positions
    VertexCTR = np.vstack([VertexC[:T, :], VertexC[-R:, :]])
    perturbation_eps = 1.0e-6  # base magnitude of perturbation in m
    perturbation_normal = np.array([-1.0, 1.0])  # set a fixed axis to perturb on
    perturbation_normal = perturbation_normal / np.sqrt(
        np.sum(perturbation_normal**2)
    )  # normalize the perturbation
    # go through the turbine/substation vertices and count how many times a
    # given vertex has appeared before
    repeat_accumulate = np.array(
        [
            int(np.sum(np.all(VertexCTR[:ivv, :] == vv, axis=1)))
            for ivv, vv in enumerate(VertexCTR)
        ]
    )
    if np.any(repeat_accumulate > 0):  # only if there are any repeats
        warn_string = (
            f"Detected {np.sum(repeat_accumulate > 0)} coincident "
            f"turbines and/or substations in optiwindnet setup."
        )  # start a warning string for the UserWarning
        # TODO: make Ard warnings?

        # create perturbation adjustements s.t. vertices w/ multiplicity > 2
        # are adjusted to be fully unique!
        adjustments = perturbation_eps * np.outer(
            repeat_accumulate, perturbation_normal
        )
        # for each adjustments add to the warning string
        for idx, dxy in enumerate(adjustments[:T, :]):
            if np.sum(dxy != 0) == 0:
                continue
            warn_string += (
                "\n\t"
                + f"adjusting turbine #{idx} from {VertexCTR[idx, :]} to  {VertexCTR[idx, :] + dxy}"
            )
        for idx, dxy in enumerate((adjustments[-R:, :])[::-1, :]):
            if np.sum(dxy != 0) == 0:
                continue
            warn_string += (
                "\n\t"
                + f"adjusting substation #{idx} from {VertexCTR[-(idx+1), :]} to {VertexCTR[-(idx+1), :] + dxy}"
            )
        # output the final warning
        warn(warn_string)

        # store the adjustments
        VertexCTR += adjustments

    # apply the adjustments
    VertexC[:T, :] = VertexCTR[:T, :]
    VertexC[-R:, :] = VertexCTR[-R:, :]

    # put together the inputs for optiwindnet
    site = dict(
        T=T,
        R=R,
        name=name_case,
        handle=name_case,
        VertexC=VertexC,
    )

    # handle the boundary if it exists
    if B > 0:
        VertexC[T:-R, 0] = discrete_inputs["x_border"]
        VertexC[T:-R, 1] = discrete_inputs["y_border"]
        site["B"] = B
        site["border"] = np.arange(T, T + B)
    return L_from_site(**site)


class OptiwindnetCollection(templates.CollectionTemplate):
    """
    Component class for modeling optiwindnet-optimized energy collection systems.

    A component class to make a heuristic-based optimized energy collection and
    management system using optiwindnet! Inherits the interface from
    `templates.CollectionTemplate`.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary

    Inputs
    ------
    x_turbines : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the turbines,
        with length `N_turbines`
    y_turbines : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the turbines,
        with length `N_turbines`
    x_substations : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the substations,
        with length `N_substations`
    y_substations : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the substations,
        with length `N_substations`

    Outputs
    -------
    total_length_cables : float
        the total length of cables used in the collection system network

    Discrete Outputs
    -------
    length_cables : np.ndarray
        a 1D numpy array that holds the lengths of each of the cables necessary
        to collect energy generated, with length `N_turbines`
    load_cables : np.ndarray
        a 1D numpy array that holds the turbine count upstream of the cable segment
        (i.e. number of turbines whose power is collected through the cable), with
        length `N_turbines`
    max_load_cables : int
        the maximum cable capacity required by the collection system
    terse_links : np.ndarray
        a 1D numpy int array encoding the electrical connections of the collection
        system (tree topology), with length `N_turbines`
    """

    def initialize(self):
        """Initialization of OM component."""
        super().initialize()
        self.S_previous: nx.Graph | None = None

    def setup(self):
        """Setup of OM component."""
        super().setup()

    def setup_partials(self):
        """Setup of OM component gradients."""

        self.declare_partials(
            ["total_length_cables"],
            ["x_turbines", "y_turbines", "x_substations", "y_substations"],
            method="exact",
        )

    def compute(
        self,
        inputs,
        outputs,
        discrete_inputs=None,
        discrete_outputs=None,
    ):
        """
        Computation for the OptiWindNet collection system design
        """

        max_turbines_per_string = self.modeling_options["collection"][
            "max_turbines_per_string"
        ]
        solver_name = self.modeling_options["collection"]["solver_name"]

        # get a graph representing the updated location
        L = _own_L_from_inputs(inputs, discrete_inputs)
        T = L.graph["T"]

        # create planar embedding and set of available links
        P, A = make_planar_embedding(L)

        solver = solver_factory(solver_name)

        model_options = self.modeling_options["collection"]["model_options"]
        # start from previous solution if available, else from heuristic if it fits
        if self.S_previous is not None:
            S_warm = self.S_previous
        elif (
            model_options.get("topology") == "branched"
            and model_options.get("feeder_limit") == "unlimited"
            and model_options.get("feeder_route") == "segmented"
        ):
            S_warm = EW_presolver(A, capacity=max_turbines_per_string)
        else:
            S_warm = None

        try:
            solver.set_problem(
                P,
                A,
                max_turbines_per_string,
                ModelOptions(**model_options),
                warmstart=S_warm,
            )
        except OWNWarmupFailed:
            # the previous solution is no longer feasible
            solver.set_problem(
                P,
                A,
                max_turbines_per_string,
                ModelOptions(**model_options),
            )

        # do the branch-and-bound MILP search
        info = solver.solve(**self.modeling_options["collection"]["solver_options"])
        S, G = solver.get_solution()
        self.S_previous = S

        # extract the outputs
        terse_links = np.zeros((T,), dtype=np.int_)
        length_cables = np.zeros((T,))
        load_cables = np.zeros((T,))

        d2roots = A.graph["d2roots"]
        # convert the graph to array representing the tree (edges i->terse[i])
        for u, v, edgeD in S.edges(data=True):
            u, v = (u, v) if u < v else (v, u)
            i, target = (u, v) if edgeD["reverse"] else (v, u)
            terse_links[i] = target
            load = edgeD["load"]
            load_cables[i] = load
            if u < 0:
                # u is a substation
                if v in G[u]:
                    # feeder <u, v> has a straight route
                    length_cables[i] = d2roots[v, u]
                else:
                    # feeder <u, v> is segmented (detoured route)
                    v_neighbors = G[v]
                    for cur_hop in v_neighbors:
                        if cur_hop >= T and v_neighbors[cur_hop]["load"] == load:
                            break
                    length_cables[i] = v_neighbors[cur_hop]["length"]
                    prev_hop = v
                    while cur_hop >= T:
                        s, t = G[cur_hop]
                        cur_hop, prev_hop = (s if t == prev_hop else t), cur_hop
                        length_cables[i] += G[cur_hop][prev_hop]["length"]
            else:
                # link (u, v) is not a feeder, so A has length data
                length_cables[i] = A[u][v]["length"]

        # pack and ship
        self.graph = G
        discrete_outputs["graph"] = G  # TODO: remove for terse links, below!
        discrete_outputs["terse_links"] = terse_links
        discrete_outputs["length_cables"] = length_cables
        discrete_outputs["load_cables"] = load_cables
        discrete_outputs["max_load_cables"] = S.graph["max_load"]
        # TODO: remove this assert after enough testing
        assert (
            abs(length_cables.sum() - G.size(weight="length")) < 1e-7
        ), f"difference: {length_cables.sum() - G.size(weight='length')}"
        outputs["total_length_cables"] = length_cables.sum()

    def compute_partials(self, inputs, J, discrete_inputs=None):

        # re-load the key variables back as locals
        G = self.graph
        T = G.graph["T"]
        R = G.graph["R"]
        VertexC = G.graph["VertexC"]
        gradients = np.zeros_like(VertexC)

        fnT = G.graph.get("fnT")
        if fnT is not None:
            _u, _v = fnT[np.array(G.edges)].T
        else:
            _u, _v = np.array(G.edges).T
        vec = VertexC[_u] - VertexC[_v]
        norm = np.hypot(*vec.T)
        # suppress the contributions of zero-length edges
        norm[np.isclose(norm, 0.0)] = 1.0
        vec /= norm[:, None]

        np.add.at(gradients, _u, vec)
        np.subtract.at(gradients, _v, vec)

        # wind turbines
        J["total_length_cables", "x_turbines"] = gradients[:T, 0]
        J["total_length_cables", "y_turbines"] = gradients[:T, 1]

        # substations
        J["total_length_cables", "x_substations"] = gradients[-R:, 0]
        J["total_length_cables", "y_substations"] = gradients[-R:, 1]

        return J



import sys
from os import PathLike
from pathlib import Path

import numpy as np
import traceback


def pyrite_validator(
    data_for_validation: dict,
    filename_pyrite: PathLike,
    rewrite: bool = False,
    rtol_val: float = 1e-6,
    load_only: bool = False,
):
    """
    TO DO!!!
    """

    # get the basename if a suffix is provided
    filename_pyrite = Path(filename_pyrite).with_suffix("")

    if rewrite:
        # this helper function can write a file to hold pyrite-standard data

        # write out a npz file that holds the variables we want to be able to check
        np.savez(filename_pyrite, **data_for_validation)
        assert False
    else:
        # ... or it can check to make sure that an existing pyrite file matches the current data

        # load an existing pyrite-standard data file
        pyrite_data = np.load(filename_pyrite.with_suffix(".npz"))

        if load_only:
            return pyrite_data

        # for each of the variables in the pyrite-standard data file
        all_validation_matches = True
        for k, v in pyrite_data.items():
            # count how many of the values in the data match the equivalent validation data
            sum_isclose = np.sum(
                np.isclose(np.array(v), data_for_validation[k], rtol=rtol_val)
            )
            vd_size = np.array(data_for_validation[k], dtype=np.float64).size
            # assert all of the values match
            validation_matches = sum_isclose == vd_size

            if not validation_matches:
                print(f"for variable {k}:", file=sys.stderr)
                prefix = "\t"
                print(
                    prefix
                    + f"{sum_isclose} values match of {vd_size} total validation values",
                    file=sys.stderr,
                )
                print(prefix + f"to a tolerance of {rtol_val:e}", file=sys.stderr)
                print(f"pyrite data for {k}: {v}", file=sys.stderr)
                print(
                    f"computed data for {k}: {data_for_validation[k]}", file=sys.stderr
                )
                stack = traceback.format_stack(limit=3)
                print("".join(stack[-2]).replace("  File", "test:"), file=sys.stderr)
                print(file=sys.stderr)

            all_validation_matches &= validation_matches
        assert all_validation_matches, "Pyrite validation data must match."



import ard


def get_stylesheets(use_tex=False, dark_background=True):
    house_style_dir = ard.BASE_DIR.parent / "assets" / "house_style"
    ard_stylesheet = (
        house_style_dir / f"stylesheet_ard{'' if use_tex else '_notex'}.mplstyle"
    )
    nrel_stylesheet = house_style_dir / f"stylesheet_nrel.mplstyle"
    styles = []
    if dark_background:
        styles.append("dark_background")
    styles.append(ard_stylesheet.as_uri())
    styles.append(nrel_stylesheet.as_uri())
    return styles



import os

import numpy as np
import matplotlib.axes
import matplotlib.pyplot as plt

import openmdao

import optiwindnet.plotting


# get plot limits based on the farm boundaries
def get_limits(
    windIOdict: dict,
    lim_buffer: float = 0.05,
):
    """
    generate plot limits based on the expected values found in the windIO file

    Parameters
    ----------
    windIOplant : dict
        a full, presumed validated, windIO plant specification file
    lim_buffer : float, optional
        a percent buffer for plot edges, by default 0.05 (5%)

    Returns
    -------
    x_lim : np.ndarray
        the two-valued limits for the x-axis based on the windIO
    y_lim : np.ndarray
        the two-valued limits for the y-axis based on the windIO
    """

    x_lim = [
        np.min(windIOdict["site"]["boundaries"]["polygons"][0]["x"])
        - lim_buffer * np.ptp(windIOdict["site"]["boundaries"]["polygons"][0]["x"]),
        np.max(windIOdict["site"]["boundaries"]["polygons"][0]["x"])
        + lim_buffer * np.ptp(windIOdict["site"]["boundaries"]["polygons"][0]["x"]),
    ]
    y_lim = [
        np.min(windIOdict["site"]["boundaries"]["polygons"][0]["y"])
        - lim_buffer * np.ptp(windIOdict["site"]["boundaries"]["polygons"][0]["y"]),
        np.max(windIOdict["site"]["boundaries"]["polygons"][0]["y"])
        + lim_buffer * np.ptp(windIOdict["site"]["boundaries"]["polygons"][0]["y"]),
    ]
    return x_lim, y_lim


def plot_layout(
    ard_prob: openmdao.api.Problem,
    input_dict: dict,
    ax: matplotlib.axes.Axes = None,
    show_image: bool = False,
    save_path: os.PathLike = None,
    save_kwargs: dict = {},
    include_cable_routing: bool = False,
    include_mooring_system: bool = False,
):
    """
    plot the layout of a farm

    Parameters
    ----------
    ard_prob : openmdao.api.Problem
        the active Ard/OpenMDAO problem
    input_dict : dict
        the active Ard input dictionary
    ax : matplotlib.axes.Axes, optional
        an already-active pyplot Axes, by default None
    show_image : bool, optional
        to show the image, rather than just saving, by default False
    save_path : os.PathLike, optional
        location where the image be saved, by default None
    save_kwargs : dict, optional
        optional keyword arguments for plt.savefig, by default {}
    include_cable_routing : bool, optional
        should the collection system routing be plotted also, by default False

    Returns
    -------
    matplotlib.axes.Axes
        the matplotlib Axes that have been generated (or modified)
    """

    # get the turbine locations to plot
    x_turbines = ard_prob.get_val("x_turbines", units="m")
    y_turbines = ard_prob.get_val("y_turbines", units="m")

    # make axis object
    if ax is None:
        fig, ax = plt.subplots()

    # plot wind plant boundaries
    windIO_dict = input_dict["modeling_options"]["windIO_plant"]

    ax.fill(
        windIO_dict["site"]["boundaries"]["polygons"][0]["x"],
        windIO_dict["site"]["boundaries"]["polygons"][0]["y"],
        linestyle="--",
        alpha=0.5,
        fill=False,
        c="k",
        # linecolor="k",
    )

    # loop over the exclusion types
    for excl_type_key, excl_type_value in (
        windIO_dict["site"].get("exclusions", {}).items()
    ):
        # handle un-implemented exclusion types
        if excl_type_key not in ["polygons"]:
            raise NotImplementedError(
                f"`{excl_type_key}` exclusions found but not yet handled in Ard plotting capabilities."
            )

        if excl_type_key == "polygons":
            for polygon_exclusion in excl_type_value:
                ax.fill(
                    polygon_exclusion["x"],
                    polygon_exclusion["y"],
                    linestyle="--",
                    alpha=0.25,
                    fill=True,
                    c="r",
                    # linecolor="k",
                )
        else:
            raise ValueError("this line should be inaccessible.")

    # plot turbines
    ax.plot(x_turbines, y_turbines, "ok")

    # adjust plot limits
    x_lim, y_lim = get_limits(windIO_dict)
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)

    if include_cable_routing:
        optiwindnet.plotting.gplot(
            ard_prob.model.collection.graph,
            ax=ax,
            dark=False,
            legend=False,
            hide_ST=True,
            infobox=False,
            landscape=False,
        )

    if include_mooring_system:
        # get the coordinates of the anchors
        x_anchors = ard_prob.get_val("x_anchors", units="m")
        y_anchors = ard_prob.get_val("y_anchors", units="m")

        # loop over the anchors and plot from their originating turbine to each
        for idx_turbine in range(
            input_dict["modeling_options"]["layout"]["N_turbines"]
        ):
            for idx_anchor in range(
                input_dict["modeling_options"]["platform"]["N_anchors"]
            ):
                ax.plot(
                    [x_turbines[idx_turbine], x_anchors[idx_turbine, idx_anchor]],
                    [y_turbines[idx_turbine], y_anchors[idx_turbine, idx_anchor]],
                    "-r",
                    alpha=0.25,
                )
            # plot the anchors as red circles
            ax.plot(
                x_anchors[idx_turbine, :],
                y_anchors[idx_turbine, :],
                "or",
                alpha=0.25,
            )

    ax.axis("equal")

    # show, save, or return
    if save_path is not None:
        plt.savefig(save_path, **save_kwargs)

    if show_image:
        plt.show()

    return ax



from pathlib import Path

from . import collection
from . import layout
from . import offshore
from . import wind_query
from . import utils
from . import viz

from .viz import house_style

BASE_DIR = Path(__file__).parent.absolute()
ASSET_DIR = BASE_DIR / "api" / "default_systems"



import numpy as np
import jax
import jax.numpy as jnp

import openmdao.api as om

import ard.utils.geometry
import ard.utils.mathematics


class MooringConstraint(om.ExplicitComponent):
    """
    A class to calculate the mooring line spacing distance for use in optimization
    constraints. Mooring lines may be defined in 2D or 3D, but the turbine positions
    are always assumed to be at sea level (z=0).

    Options
    -------
    modeling_options : dict
        a modeling options dictionary (inherited from `FarmAeroTemplate`)
    wind_query : floris.wind_data.WindRose
        a WindQuery objects that specifies the wind conditions that are to be
        computed
    bathymetry_data : ard.geographic.BathymetryData
        a BathymetryData object to specify the bathymetry mesh/sampling

    Inputs
    ------
    x_turbines : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the turbines,
        with length `N_turbines` (mirrored w.r.t. `FarmAeroTemplate`)
    y_turbines : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the turbines,
        with length `N_turbines` (mirrored w.r.t. `FarmAeroTemplate`)
    x_anchors : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the mooring
        system anchors, with shape `N_turbines` x `N_anchors`
    y_anchors : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the mooring
        system anchors, with shape `N_turbines` x `N_anchors`
    z_anchors (optional) : np.ndarray
        a 1D numpy array indicating the z-dimension locations of the mooring
        system anchors, with shape `N_turbines` x `N_anchors`
    """

    def initialize(self):
        """Initialization of the OpenMDAO component."""
        self.options.declare("modeling_options")

    def setup(self):
        """Setup of the OpenMDAO component."""

        # load modeling options
        self.modeling_options = self.options["modeling_options"]
        self.N_turbines = int(self.modeling_options["layout"]["N_turbines"])
        self.N_anchor_dimensions = int(
            self.modeling_options["platform"]["N_anchor_dimensions"]
        )
        self.N_anchors = int(self.modeling_options["platform"]["N_anchors"])
        self.N_distances = int((self.N_turbines - 1) * self.N_turbines / 2)

        # set up inputs and outputs for mooring system
        self.add_input(
            "x_turbines", jnp.zeros((self.N_turbines,)), units="km"
        )  # x location of the mooring platform in km w.r.t. reference coordinates
        self.add_input(
            "y_turbines", jnp.zeros((self.N_turbines,)), units="km"
        )  # y location of the mooring platform in km w.r.t. reference coordinates
        self.add_input(
            "x_anchors",
            jnp.zeros((self.N_turbines, self.N_anchors)),
            units="km",
        )  # x location of the mooring anchors in km w.r.t. reference coordinates
        self.add_input(
            "y_anchors",
            jnp.zeros((self.N_turbines, self.N_anchors)),
            units="km",
        )  # y location of the mooring anchors in km w.r.t. reference coordinates
        if self.N_anchor_dimensions == 3:
            self.add_input(
                "z_anchors",
                jnp.zeros((self.N_turbines, self.N_anchors)),
                units="km",
            )  # z location of the mooring anchors in km w.r.t. reference coordinates

        self.add_output(
            "mooring_spacing",
            jnp.zeros(self.N_distances),
            units="km",
        )  # consolidated violation length

    def setup_partials(self):
        """Derivative setup for the OpenMDAO component."""
        # the default (but not preferred!) derivatives are FDM
        self.declare_partials("*", "*", method="exact")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """Computation for the OpenMDAO component."""

        # unpack the working variables
        x_turbines = inputs["x_turbines"]
        y_turbines = inputs["y_turbines"]
        x_anchors = inputs["x_anchors"]
        y_anchors = inputs["y_anchors"]
        if self.N_anchor_dimensions == 3:
            z_anchors = inputs["z_anchors"]

        if self.N_anchor_dimensions == 2:
            distances = mooring_constraint_xy(
                x_turbines, y_turbines, x_anchors, y_anchors
            )
        elif self.N_anchor_dimensions == 3:
            distances = mooring_constraint_xyz(
                x_turbines, y_turbines, x_anchors, y_anchors, z_anchors
            )
        else:
            raise (ValueError("modeling_options['layout'][']"))

        outputs["mooring_spacing"] = distances

    def compute_partials(self, inputs, partials, discrete_inputs=None):

        # unpack the working variables
        x_turbines = inputs["x_turbines"]
        y_turbines = inputs["y_turbines"]
        x_anchors = inputs["x_anchors"]
        y_anchors = inputs["y_anchors"]
        if self.N_anchor_dimensions == 3:
            z_anchors = inputs["z_anchors"]

        if self.N_anchor_dimensions == 2:
            jacobian = mooring_constraint_xy_jac(
                x_turbines, y_turbines, x_anchors, y_anchors
            )
        elif self.N_anchor_dimensions == 3:
            jacobian = mooring_constraint_xyz_jac(
                x_turbines, y_turbines, x_anchors, y_anchors, z_anchors
            )
        else:
            raise (ValueError("modeling_options['layout'][']"))

        partials["mooring_spacing", "x_turbines"] = jacobian[0]
        partials["mooring_spacing", "y_turbines"] = jacobian[1]
        partials["mooring_spacing", "x_anchors"] = jacobian[2]
        partials["mooring_spacing", "y_anchors"] = jacobian[3]
        if self.N_anchor_dimensions == 3:
            partials["mooring_spacing", "z_anchors"] = jacobian[4]


def mooring_constraint_xy(
    x_turbines: np.ndarray,
    y_turbines: np.ndarray,
    x_anchors: np.ndarray,
    y_anchors: np.ndarray,
):
    """Mooring distance calculation in 2 dimensions

    Args:
        x_turbines (np.ndarray): array of turbine x positions
        y_turbines (np.ndarray): array of turbine y positions
        x_anchors (np.ndarray): array of anchor x positions
        y_anchors (np.ndarray): array of anchor y positions

    Returns:
        np.ndarray: 1D array of distances with length (n_turbines - 1)*n_turbines/2
    """

    # convert inputs
    mooring_points = convert_inputs_x_y_to_xy(
        x_turbines, y_turbines, x_anchors, y_anchors
    )
    # calculate minimum distances between each set of moorings
    distances = calc_mooring_distances(mooring_points)

    return distances


mooring_constraint_xy = jax.jit(mooring_constraint_xy)
mooring_constraint_xy_jac = jax.jacrev(mooring_constraint_xy, argnums=[0, 1, 2, 3])


def mooring_constraint_xyz(
    x_turbines: np.ndarray,
    y_turbines: np.ndarray,
    x_anchors: np.ndarray,
    y_anchors: np.ndarray,
    z_anchors: np.ndarray,
):
    """Mooring distance calculation in 3 dimensions. The third dimension
    is only required for the anchors since the turbine platforms are
    all assumed to be at sea level.

    Args:
        x_turbines (np.ndarray): array of turbine x positions
        y_turbines (np.ndarray): array of turbine y positions
        x_anchors (np.ndarray): array of anchor x positions
        y_anchors (np.ndarray): array of anchor y positions
        z_anchors (np.ndarray): array of anchor z positions

    Returns:
        np.ndarray: 1D array of distances with length (n_turbines - 1)*n_turbines/2
    """

    # convert inputs
    mooring_points = convert_inputs_x_y_z_to_xyz(
        x_turbines,
        y_turbines,
        np.zeros(len(x_turbines)),
        x_anchors,
        y_anchors,
        z_anchors,
    )
    # calculate minimum distances between each set of moorings
    distances = calc_mooring_distances(mooring_points)

    return distances


mooring_constraint_xyz = jax.jit(mooring_constraint_xyz)
mooring_constraint_xyz_jac = jax.jacrev(mooring_constraint_xyz, argnums=[0, 1, 2, 3, 4])


def calc_mooring_distances(mooring_points: np.ndarray) -> np.ndarray:
    """Calculate the minimum distances between each set of mooring lines

    Args:
        mooring_points (np.ndarray): array of mooring points of shape
        (n_turbines, n_anchors+1, n_dimensions) where n_dimensions may be 2 or 3

    Returns:
        np.ndarray: 1D array of distances with length (n_turbines - 1)*n_turbines/2
    """

    n_turbines = mooring_points.shape[0]

    # Create pairwise indices for all unique turbine pairs
    i_indices, j_indices = jnp.triu_indices(n_turbines, k=1)

    # Extract the corresponding mooring points for each pair
    mooring_A = mooring_points[i_indices]
    mooring_B = mooring_points[j_indices]

    # Compute distances for all pairs using `distance_mooring_to_mooring`
    distances = jax.vmap(distance_mooring_to_mooring)(mooring_A, mooring_B)

    return distances


def convert_inputs_x_y_to_xy(
    x_turbines: np.ndarray,
    y_turbines: np.ndarray,
    x_anchors: np.ndarray,
    y_anchors: np.ndarray,
) -> np.ndarray:
    """Convert from inputs of x for turbines, y for turbines, x for anchors, and y for
    anchors to single array for mooring specification that is of shape
    (n_turbines, n_anchors+1, 2). for each set of points, the turbine position is given
    first followed by the anchor positions.

    Args:
        x_turbines (np.ndarray): array of turbine x positions
        y_turbines (np.ndarray): array of turbine y positions
        x_anchors (np.ndarray): array of anchor x positions
        y_anchors (np.ndarray): array of anchor y positions

    Returns:
        np.ndarray: all turbine and anchor location information combined into a single
            array of shape (n_turbines, n_anchors+1, 2)
    """

    # Stack turbine positions and anchor positions directly
    turbine_positions = jnp.stack([x_turbines, y_turbines], axis=-1)[:, None, :]
    anchor_positions = jnp.stack([x_anchors, y_anchors], axis=-1)

    # Concatenate turbine positions with anchor positions
    xy = jnp.concatenate([turbine_positions, anchor_positions], axis=1)

    return xy


def convert_inputs_x_y_z_to_xyz(
    x_turbines: np.ndarray,
    y_turbines: np.ndarray,
    z_turbines: np.ndarray,
    x_anchors: np.ndarray,
    y_anchors: np.ndarray,
    z_anchors: np.ndarray,
) -> np.ndarray:
    """Convert from inputs of x for turbines, y for turbines, z for turbines, x for anchors,
    y for anchors, and z for anchors to single array for mooring specification that is of
    shape (n_turbines, n_anchors+1, 3). for each set of points, the turbine position is given
    first followed by the anchor positions.

    Args:
        x_turbines (np.ndarray): array of turbine x positions
        y_turbines (np.ndarray): array of turbine y positions
        z_turbines (np.ndarray): array of turbine z positions
        x_anchors (np.ndarray): array of anchor x positions
        y_anchors (np.ndarray): array of anchor y positions
        z_anchors (np.ndarray): array of anchor z positions

    Returns:
        np.ndarray: all input information combined into a single array of shape (n_turbines, n_anchors+1, 3)
    """

    # Stack turbine positions and anchor positions directly
    turbine_positions = jnp.stack([x_turbines, y_turbines, z_turbines], axis=-1)[
        :, None, :
    ]
    anchor_positions = jnp.stack([x_anchors, y_anchors, z_anchors], axis=-1)

    # Concatenate turbine positions with anchor positions
    xyz = jnp.concatenate([turbine_positions, anchor_positions], axis=1)

    return xyz


def distance_point_to_mooring(point: np.ndarray, P_mooring: np.ndarray) -> float:
    """Find the distance from a point to a set of mooring lines for a single floating wind turbine.
        While arguments may be given in either 2d ([x,y]) or 3d ([x,y,z]), the point of interest
        and the mooring line points must all be given with the same number of dimensions.

    Args:
        point (np.ndarray): Point of interest in 2d ([x,y]) or 3d ([x,y,z]).
        P_mooring (np.ndarray): The set of points defining the mooring line layout. The first point should
                                be the center, the rest of the points define the anchor points. Points may
                                be given in 2d ([x,y]) or 3d ([x,y,z]).

    Returns:
        float: The shortest distance from the point of interest to the set of mooring lines.
    """

    p_center = P_mooring[0]
    anchors = P_mooring[1:]

    distances = jax.vmap(
        ard.utils.geometry.distance_point_to_lineseg_nd,
        in_axes=(None, None, 0),
    )(point, p_center, anchors)

    return ard.utils.mathematics.smooth_min(distances)


def distance_mooring_to_mooring(
    P_mooring_A: np.ndarray, P_mooring_B: np.ndarray
) -> float:
    """Calculate the distance from one set of mooring lines to another. Moorings
    are defined with the center point (platform location) first, followed by the
    anchor points in no specific order.

    Args:
        P_mooring_A (np.ndarray): ndarray of points of mooring A of shape (npoints, nd) (e.g. (4, (x, y, z))).
            Center point must come first.
        P_mooring_B (np.ndarray): ndarray of points of mooring B of shape (npoints, nd) (e.g. (4, (x, y, z))).
            Center point must come first.

    Returns:
        float: shortest distance between the two sets of moorings
    """

    p_center_A = P_mooring_A[0]
    p_center_B = P_mooring_B[0]
    anchors_A = P_mooring_A[1:]
    anchors_B = P_mooring_B[1:]

    # Vectorize the computation of distances between all pairs of line segments
    def compute_segment_distance(anchor_A, anchor_B):
        return ard.utils.geometry.distance_lineseg_to_lineseg_nd(
            p_center_A, anchor_A, p_center_B, anchor_B
        )

    # Use vmap to compute distances for all combinations of anchors
    distances = jax.vmap(
        lambda anchor_A: jax.vmap(
            lambda anchor_B: compute_segment_distance(anchor_A, anchor_B)
        )(anchors_B)
    )(anchors_A)

    # Find the smooth minimum distance
    return ard.utils.mathematics.smooth_min(distances.flatten())
